import { AuthState } from "./enum/AuthState";
import { Endian, createReader, createWriter } from "bufferstuff";
import { existsSync, readFileSync } from "fs";
import { createServer, Socket } from "net";
import { Packet } from "./enum/Packet";
import { WebSocket } from "ws";
import IClientConfig from "./interface/IClientConfig";
import Constants from "./Constants";

export default class Client {
	constructor() {
		if (!existsSync("./config/client-config.json")) {
			console.error("client-config.json is missing!");
			process.exit(1);
		}
		const config:IClientConfig = JSON.parse(readFileSync("./config/client-config.json").toString());
		let CLIENT_ID = 0;
		const server = createServer((socket) => {
			let authed = false;
			let queuedMessages = new Array<Buffer>();
			let txBytes = 0;
			let rxBytes = 0;
			const thisClientId = CLIENT_ID++;
			console.log(`[LOCAL CLIENT ${thisClientId}] New Connection`);
			const txrxInterval = setInterval(() => {
				console.log(`[LOCAL CLIENT ${thisClientId}] TX: ${(txBytes / 1024).toFixed(2)}KB/s  |  RX: ${(rxBytes / 1024).toFixed(2)}KB/s`);
				txBytes = rxBytes = 0;
			}, 1000);
	
			const client = new Socket();
			client.on("connect", () => {
				// Send off auth as soon as we connect
				client.write(createWriter(Endian.LE, 2 + config.authKey.length).writeUByte(Packet.Auth).writeShortString(config.authKey).toBuffer());
			});
	
			client.on("data", (data) => {
				// @ts-ignore
				const packetData = createReader(Endian.LE, data);
				const packetId = packetData.readUByte();
				let tempReadLength = 0;
				let bufferData:Buffer;
	
				switch (packetId) {
					case Packet.KeepAlive:
						client.write(Constants.KEEPALIVE_PACKET);
						break;
	
					case Packet.Auth:
						if (authed) {
							return;
						}
	
						authed = packetData.readUByte() === AuthState.Good;
						if (authed && queuedMessages.length > 0) {
							for (const message of queuedMessages) {
								client.write(message);
							}
							queuedMessages.length = 0;
						}
						break;
					case Packet.Data:
						tempReadLength = packetData.readUInt();
						bufferData = packetData.readBuffer(tempReadLength);
						rxBytes += bufferData.length;
						socket.write(bufferData);
						break;
				}
			});
	
			function clientCloseOrError() {
				socket.end();
				clearInterval(txrxInterval);
				console.log(`[LOCAL CLIENT ${thisClientId}] Remote server terminated the connection.`);
			}
			client.on("close", clientCloseOrError);
			client.on("error", clientCloseOrError);
	
			socket.on("data", (chunk) => {
				const bufferData = createWriter(Endian.LE, 5).writeUByte(Packet.Data).writeUInt(chunk.length).writeBuffer(chunk).toBuffer();
				txBytes += chunk.length;
				if (authed) {
					client.write(bufferData);
				} else {
					queuedMessages.push(bufferData);
				}
			});
	
			function serverCloseOrError() {
				client.end();
				clearInterval(txrxInterval);
				console.log(`[LOCAL CLIENT ${thisClientId}] Disconnected`);
			}
			socket.on("close", serverCloseOrError);
			socket.on("error", serverCloseOrError);

			client.connect({
				host: config.remoteHost,
				port: config.remotePort
			});
		});
	
		server.listen(config.localPort, () => console.log(`Local server listening at ${config.localPort}`));
	}
}